import {
  Class,
  Method,
} from '@overflow/commons/core/reflect';
import { ClassType, PropertyKeyType } from '@overflow/commons/core/type';
import { ApplicationContext } from './context';
import {
  ConfigurationAnnotation,
  InstanceAnnotation,
  RunnerAnnotation,
  WebApplicationAnnotation,
} from './decorators';
import { InstanceDefinition } from '@overflow/commons/di/factory';

export class Application {
  private _primaryClass: ClassType;
  private _applicationContext: ApplicationContext;

  public static async run(primaryClass: ClassType): Promise<ApplicationContext> {
    let app: Application = new Application(primaryClass);
    return app.run();
  }

  public constructor(primaryClass: ClassType) {
    this._primaryClass = primaryClass;

    this.createContext();
  }

  /**
   * run
   */
  public async run(): Promise<ApplicationContext> {
    let runner = this.findRunner();
    if (undefined === runner) {
      throw new Error(`There is not exist @Runner on Application[${this._primaryClass.name}]`);
    }
    let instanceFactory = this._applicationContext.instanceFactory;

    let clazz: Class = Class.forClass(this._primaryClass);
    let instance = clazz.getConstructor().newInstance();
    instanceFactory.applyInstance(this._primaryClass, instance);

    await runner.invoke(instance);

    return this._applicationContext;
  }

  private findRunner(): Method | undefined {
    let clazz: Class = Class.forClass(this._primaryClass);
    let methods = clazz.getOwnMethods();

    for (let key of Array.from(methods.keys())) {
      let method = methods.get(key);
      let runnerAnnotation = method.getOwnAnnotation(RunnerAnnotation);
      if (undefined === runnerAnnotation) {
        continue;
      }
      return method;
    }

    return undefined;
  }

  private createContext(): ApplicationContext {
    let clazz: Class = Class.forClass(this._primaryClass);

    let wa: WebApplicationAnnotation = clazz.getOwnAnnotation(WebApplicationAnnotation);
    if (undefined === wa) {
      throw new Error(`Class is not WebApplication type. add @WebApplication annotation to class[${this._primaryClass.name}]`);
    }

    this._applicationContext = new ApplicationContext(wa.attributes.injectables);

    this.registerJSONSources(wa.attributes.jsonSources);
    this.registerConfigurations(wa.attributes.configurations);

    return this._applicationContext;
  }

  private registerJSONSources(jsonSources: any[]): void {
    if (undefined === jsonSources || null === jsonSources) {
      return;
    }
    let instanceFactory = this._applicationContext.instanceFactory;
    let jsons = Object.assign({}, ...jsonSources);
    instanceFactory.registerJSON(jsons);
  }

  private registerConfigurations(configurationTypes: ClassType[]): void {
    if (undefined === configurationTypes || null === configurationTypes) {
      return;
    }
    let instanceFactory = this._applicationContext.instanceFactory;

    configurationTypes.forEach(configurationType => {
      let clazz: Class = Class.forClass(configurationType);
      let configAnnotation: ConfigurationAnnotation = clazz.getOwnAnnotation(ConfigurationAnnotation);
      if (undefined === configAnnotation) {
        throw new Error(`Class is not Configuration type. add @Configuration annotation to class[${configurationType.name}]`);
      }
      let methods = clazz.getOwnMethods();
      if (null === methods) {
        return;
      }
      let instance = clazz.getConstructor().newInstance();
      instanceFactory.applyInstance(configurationType, instance);
      methods.forEach(method => {
        let instanceAnnotation = method.getOwnAnnotation(InstanceAnnotation);
        if (undefined === instanceAnnotation) {
          return;
        }
        let singletonName = null === instanceAnnotation.attributes.name ? method.getName() : instanceAnnotation.attributes.name;
        let result = method.invoke(instance);
        instanceFactory.registerSingleton(singletonName, method.getReturnType(), result);
      });
    });
  }
}
